﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.9
@EndOfDesignText@
'Module Class: DeviceStateClass
Sub Class_Globals
	' XML-API Scripts- MUST START WITH /
	Public SCRIPT_DEVICELIST As String = "/addons/xmlapi/devicelist.cgi"
	Public SCRIPT_STATELIST As String = "/addons/xmlapi/statelist.cgi"
	Public SCRIPT_SYSVARLIST As String = "/addons/xmlapi/sysvarlist.cgi"

	' Set filenames for saving xmlapi response
	Private FILE_DEVICELIST As String = "devicelist.xml"
	Private FILE_STATELIST As String = "statelist.xml"
	Private FILE_SYSVARLIST As String = "sysvarlist.xml"

	' Parsed statelist from the XMLAPI HTTP Response
	Private Statelist As List
	
	' List of devices found in the parsed statelist - used for the customlistview devices
	' The devicelist contains an object with an array holding attributes device name and device ise_id
	' Example: DeviceList.Add(Array(attributes.Get(ATTRIBUTE_NAME), attributes.Get(ATTRIBUTE_ISE_ID)))
	Public DeviceList As List

	' List of sysvars
	Public SysvarList As List

	' Build a list of all type attribute from the XMLAPI HTTP Response tree node <datapoints ...>
	' Example type CONFIG_PENDING: <datapoint name="HmIP-RF.002018A99D2097:0.CONFIG_PENDING" Type="CONFIG_PENDING" ...
	Public TypeAttributes As List

	' XML-API Device Attributes - IMPORTANT: Use the right case (check the XML-API request result)
	Public Const ATTRIBUTE_ADDRESS As String			= "address"
	Public Const ATTRIBUTE_CHANNEL As String			= "channel"
	Public Const ATTRIBUTE_DATAPOINT As String			= "datapoint"
	Public Const ATTRIBUTE_DEVICE As String				= "device"
	Public Const ATTRIBUTE_DEVICELIST As String			= "deviceList"
	Public Const ATTRIBUTE_DEVICE_TYPE As String		= "device_type"
	Public Const ATTRIBUTE_ISE_ID As String				= "ise_id"
	Public Const ATTRIBUTE_NAME As String				= "name"
	Public Const ATTRIBUTE_STATELIST As String			= "stateList"
	Public Const ATTRIBUTE_TYPE As String				= "type"
	Public Const ATTRIBUTE_VALUE As String				= "value"
	Public Const ATTRIBUTE_SYSTEMVARIABLES As String	= "systemVariables"
	Public Const ATTRIBUTE_SYSTEMVARIABLE As String		= "systemVariable"

	Public Const KEY_ATTRIBUTES As String				= "Attributes"

	Public FILTER_TYPE_DEFAULT As String				=  "ACTUAL_TEMPERATURE"
	Public FILTER_TYPE_LOW_BAT As String				=  "LOW_BAT"
	' Public FILTER_TYPE() As String =  Array As String("ACTUAL_TEMPERATURE", "LOW_BAT", "STATE", "SET_POINT_TEMPERATURE")
End Sub

' Initialize the object - the lists used
Public Sub Initialize
	DeviceList.Initialize
	Statelist.Initialize
	TypeAttributes.Initialize
	SysvarList.Initialize
End Sub

#Region PARSEDEVICELIST
' Parse the devicelist XML to a MAP
' Retrieve the HomeMatic CCU statelist via XML-API URL request.
' Example: http://192.168.1.70/addons/xmlapi/devicelist.cgi or http://ccu3-webui/addons/xmlapi/devicelist.cgi
' The result is an XML tree with all devices & datapoints.
' Parsed are the attributes: device name, ise_id, device_type
' <DeviceList>
'	<device name="Bad" address="002018A99D2097" ise_id="1786" interface="HmIP-RF" device_type="HmIP-eTRV-B" ready_config="true">
'	</device>
'	<device name="Briefkasten" address="0000DA498D5859" ise_id="2508" interface="HmIP-RF" device_type="HMIP-SWDO" ready_config="true">
' ...
Public Sub Parse_Devicelist(xmlData As String, xmlSave As Boolean)
	Dim xm As Xml2Map			'Lib Xml2Map
	Dim ParsedData As Map		'Holding the parsed xml tree as a map
	' Option to save the retrieved statelist to a textfile in the dirapp folder
	' The xmlapi response is a one liner, add CRLF to make more readible
	If xmlSave Then
		File.WriteString(File.DirApp, FILE_DEVICELIST, xmlData.Replace(">", $">${CRLF}"$))
	End If

	' Clear the lists holding the objects for the devices
	DeviceList.Clear

	' Init the object used to parse the xml data (string)
	xm.Initialize

	' Parse the data to a map
	ParsedData = xm.Parse(xmlData)
	' Get the DeviceList attribute (ensure the right case) as a map. It holds a single entry "DeviceList"
	Dim mapDeviceList As Map = ParsedData.Get(ATTRIBUTE_DEVICELIST)
	' Check if the map contains data
	If Not(mapDeviceList.IsInitialized) Then
		Return
	End If

	' OPTION
	' Save to file to analyse map structure - only required during development
	' File.WriteMap(File.DirApp, "devicelist.map", mapDeviceList)

	' Get the devices as list from the map mapDeviceList single entry. Each device is a map in the mapDeviceList.
	' device= [{Attributes={name=Bad, ise_id=1786, unreach=False, config_pending=False},
	Dim TempDeviceList As List = mapDeviceList.Get(ATTRIBUTE_DEVICE)
	' Log($"Devices found: ${Statelist.Size}"$)	'Devicesfound: 14

	' Loop over the statelist to get the device as map and add to the devices selection list (delivelist used for the customlistview devices)
	' The devicelist contains an object with an array holding the 4 device attributes: name, ise_id, device_type, address
	' To get data from the list (l) use a for each loop like:
	'	For Each o() As Object In l
	'		Dim objName As String = o(0)
	'		Dim objIseId As String = o(1)
	'		Dim objDeviceType As String = o(2)
	'		Dim objAddress As String = o(3)
	'	Next
	For Each device As Map In TempDeviceList
		' Get the attributes for a device as a map using "Attribute"
		' Example: Attributes={name=Bad, ise_id=1786, unreach=False, config_pending=False},
		Dim attributes As Map = device.Get(KEY_ATTRIBUTES)
		' Log($"${attributes.Get(ATTRIBUTE_NAME)}, ${attributes.Get(ATTRIBUTE_ISE_ID)}, ${attributes.Get(ATTRIBUTE_DEVICE_TYPE)}"$)
		DeviceList.Add(Array( _ 
			attributes.Get(ATTRIBUTE_NAME), _ 
			attributes.Get(ATTRIBUTE_ISE_ID), _ 
			attributes.Get(ATTRIBUTE_DEVICE_TYPE), _ 
			attributes.Get(ATTRIBUTE_ADDRESS)))
	Next
End Sub
#End Region

#Region PARSESTATELIST
'Parse the statelist XML to a MAP
'Retrieve the HomeMatic CCU statelist via XML-API URL request.
'Example: http://192.168.1.70/addons/xmlapi/statelist.cgi or http://ccu3-webui/addons/xmlapi/statelist.cgi
'The result is an XML tree with all devices & datapoints
'<stateList>
'	<device name="Bad" ise_id="1786" unreach="false" config_pending="false">
'		<channel name="Bad:0" ise_id="1787" index="0" visible="true" operate="true">
'			<datapoint name="HmIP-RF.002018A99D2097:0.CONFIG_PENDING" Type="CONFIG_PENDING" ise_id="1788" value="false" valuetype="2" valueunit="" timestamp="1613728140" operations="5"/>
'			<datapoint name="HmIP-RF.002018A99D2097:0.DUTY_CYCLE" Type="DUTY_CYCLE" ise_id="1792" value="false" valuetype="2" valueunit="" timestamp="1613728140" operations="5"/>
' ...
Public Sub Parse_Statelist(xmlData As String, xmlSave As Boolean)
	Dim xm As Xml2Map			'Lib Xml2Map
	Dim ParsedData As Map		'Holding the parsed xml tree as a map

	' Option to save the retrieved statelist to a textfile in the dirapp folder
	' The xmlapi response is a one liner, add CRLF to make more readible
	If xmlSave Then
		File.WriteString(File.DirApp, FILE_STATELIST, xmlData.Replace(">", $">${CRLF}"$))
	End If

	' Get all type attribute (i.e. type="LOW_BAT") used to pick from for the filter (see main menu item: filter > select)
	TypeAttributes = Get_All_Type_Attributes(xmlData)
	
	' Clear the lists holding the objects for the devices and the statelist
	Statelist.Clear

	' Init the object used to parse the xml data (string)
	xm.Initialize

	' Parse the data to a map
	ParsedData = xm.Parse(xmlData)
	' Get the stateList attribute (ensure the right case) as a map. It holds a single entry "stateList"
	Dim mapStateList As Map = ParsedData.Get(ATTRIBUTE_STATELIST)
	' Check if the map contains data
	If Not(mapStateList.IsInitialized) Then
		Return
	End If

	' OPTION
	' Save to file to analyse map structure - only required during development
	' File.WriteMap(File.DirApp, "statelist.map", mapStateList)

	' Get the devices as list from the map statelist single entry. Each device is a map in the statelist.
	' device= [{Attributes={name=Bad, ise_id=1786, unreach=False, config_pending=False},
	Statelist = mapStateList.Get(ATTRIBUTE_DEVICE)
	' Log($"Devices found: ${Statelist.Size}"$)	'Devicesfound: 14
End Sub
#End Region

' Get the device attributes from the global list DeviceList for a device given id (using attribute ise_id)
Private Sub Get_Device_Attributes(ID As Int) As Map
	' Log($"Get_Device_Attributes: ${ID}"$)
	For Each deviceItem As Map In DeviceList
		' Get the attributes for a device as a map
		Dim deviceAttributes As Map = deviceItem.Get(KEY_ATTRIBUTES)
		' Check if the device ise_id maches the ID argument
		Dim ise_id As Int = deviceItem.Get("ise_id")
		If ise_id = ID Then
			Return deviceAttributes
		End If
	Next
	Return Null
End Sub

' Get the datapoints for a device with ID ("ise_id") and optional TypeFilter
' The datapoints are added to a list as object holding an array of 4 datapoint attributes
' Example: 
' dpItemList.Add(Array(dpAttribute.Get(ATTRIBUTE_TYPE), dpAttribute.Get(ATTRIBUTE_VALUE), dpAttribute.Get(ATTRIBUTE_ISE_ID), channelName))
' To get data from the list (l) use a for each loop like:
'	For Each o() As Object In l
'		Dim objType As String = o(0)
'		Dim objValue As String = o(1)
'		Dim objIseId As String = o(2)
'		Dim objChannel As String = o(3)
'	Next
' The TypeFilter enables to select a specific type (uppercase), i.e. ACTUAL_TEMPERATURE.
' This can be used to select for all devices a specific type, like for thermostats HmIP-eTRV...
' Return List
Public Sub Get_Device_Datapoints(ID As Int, TypeFilter As String, Export As Boolean) As List
	' Log($"Get_Device_Datapoints: ${ID}"$)
	' Search for the device datapoints by looping over the devices list and add to the listview devices
	' Define the list with datapoints
	Dim dpItemList As List
	dpItemList.Initialize
	' If there is a TypeFilder, ensure to uppercase
	If TypeFilter.Length > 0 Then
		TypeFilter = TypeFilter.ToUpperCase
	End If
	' Loop over the device items
	For Each deviceItem As Map In Statelist
		' Get the attributes for a device as a map
		Dim deviceAttributes As Map = deviceItem.Get(KEY_ATTRIBUTES)
		' Check if the device ise_id matches the ID (Value argument)
		Dim ise_id As Int = deviceAttributes.Get(ATTRIBUTE_ISE_ID)
		If ise_id = ID Then
			' Get the device channel(s) as list
			Dim deviceChannels As List = deviceItem.Get(ATTRIBUTE_CHANNEL)
			' Loop over the device channels to get the datapoints
			For Each ch As Map In deviceChannels
				Dim chAttribute As Map = ch.Get(KEY_ATTRIBUTES)
				' Get the name of the channel
				Dim channelName As String = "Unknown"
				If chAttribute.ContainsKey(ATTRIBUTE_NAME) Then
					channelName = chAttribute.Get(ATTRIBUTE_NAME)
				End If
				' Check if there is a key holding datapoint(s)
				' IMPORTANT
				' If single datapoint then ch.get("datapoint") is a map
				' If multiple datapoints then ch.get("datapoint") is a list
				If ch.ContainsKey(ATTRIBUTE_DATAPOINT) Then
					' Get the datapoint as an object and determine if map (single entry) or list (multiple entries)
					Dim dpData As Object = ch.Get(ATTRIBUTE_DATAPOINT)
					' Create the list holding the datapoint(s) map
					Dim dpList As List
					dpList.Initialize
					' If object is map add map to the datapoint list
					If dpData Is Map Then
						dpList.Add(dpData)
					End If
					' If object is a list assign the object as list to the datapoint list
					If dpData Is List Then
						dpList = dpData
					End If
					' Loop over the device datapoint list holding a map for each of the datapoints
					For Each dp As Map In dpList
						' Get the datapoint "Attribute" to select attributes "type", "value", "ise_id" or more if required
						' Example (MyMap) {name=HmIP-RF.002018A99D2097:1.ACTUAL_TEMPERATURE, type=ACTUAL_TEMPERATURE, ise_id=1808, value=23.000000, valuetype=4, valueunit=, timestamp=1613814228, operations=5}
						Dim dpAttribute As Map = dp.Get(KEY_ATTRIBUTES)
						' Add item data: Check for Export or if there is a TypeFilter and Type matches or if there is no TypeFilter
						If Export Then
							' Iterate over the attribute map keys and values. The string contains all attributes.
							Dim s As String
							For Each ItemKey As String In dpAttribute.Keys
								Dim ItemValue As String = dpAttribute.Get(ItemKey)
								s = s & $"${ItemKey} = ${ItemValue}"$ & ","
							Next
							' Add the device attribute as a single line to the list
							dpItemList.Add(s.SubString2(0, s.LastIndexOf(",")))
						End If
						If Not(Export) Then
							If (TypeFilter.Length > 0 And dpAttribute.Get(ATTRIBUTE_TYPE) == TypeFilter) Or (TypeFilter.Length == 0) Then
								' Add the item data as text to the list.
								dpItemList.Add(Array( _
									dpAttribute.Get(ATTRIBUTE_TYPE), _ 
									dpAttribute.Get(ATTRIBUTE_VALUE), _ 
									dpAttribute.Get(ATTRIBUTE_ISE_ID), _ 
									channelName))
							End If							
						End If
					Next
				End If
			Next
		End If
	Next
	Return dpItemList
End Sub

' Get all attribute types from the xmldata
' Example: get type='CONFIG_PENDING' from <datapoint name='HmIP-RF.002018A99D2097:0.CONFIG_PENDING' type='CONFIG_PENDING' ise_id='1788' value='false' valuetype='2' valueunit='' timestamp='1614067920' operations='5' />
' Returns List
Private Sub Get_All_Type_Attributes(xmldata As String) As List
	' Set pattens to find [ type='MYTYPE'] - note the blank prior type else other attributes line valuetype='2' are also found
	Dim pattern As String = " type='([^']*)'"	' "/'(.*?)'/"
	Dim Typelist As List
	Typelist.Initialize
	Dim Matcher1 As Matcher
	Matcher1 = Regex.Matcher(pattern, xmldata)
	Do While Matcher1.Find
		Dim TypeItem As String = Matcher1.Match.Replace(" type=","").Replace("'","")
		If Typelist.IndexOf(TypeItem) = - 1 Then Typelist.Add(TypeItem)
		' Log("Found: " & TypeItem)
	Loop
	Typelist.Sort(True)
	Return Typelist
End Sub

#Region PARSESTATELIST_WITH_DEVICELIST
' Parse the statelist XML to a MAP and build the devicelist with 2 items name & ised_id
' NOT USED because the device_type is not in the xmlapi statelist result which is used in the devivelist customlistview
Private Sub Parse_Statelist_With_DeviceList(xmlData As String, xmlSave As Boolean)	'ignore
	Dim xm As Xml2Map			'Lib Xml2Map
	Dim ParsedData As Map

	' Option to save the statelist to a textfile in the dirapp folder
	If xmlSave Then
		File.WriteString(File.DirApp, FILE_STATELIST, xmlData.Replace(">", $">${CRLF}"$))
	End If

	' Get all attribute types used to pick from for the filter
	TypeAttributes = Get_All_Type_Attributes(xmlData)
	
	' Clear the lists holding the objects
	DeviceList.Clear
	Statelist.Clear
	' Init the object used to parse the xml data (string)
	xm.Initialize
	' Parse the data to a map
	ParsedData = xm.Parse(xmlData)
	' Get the stateList attribute (ensure the right case) as a map. It holds a single entry "stateList"
	Dim mapStateList As Map = ParsedData.Get(ATTRIBUTE_STATELIST)
	' Check if the map contains data
	If Not(mapStateList.IsInitialized) Then
		Return
	End If

	' OPTION
	' Save to file to analyse map structure - only required during development
	' File.WriteMap(File.DirApp, "statelist.map", mapStateList)

	' Get the devices as list from the map statelist single entry. Each device is a map in the statelist.
	' device= [{Attributes={name=Bad, ise_id=1786, unreach=False, config_pending=False},
	Statelist = mapStateList.Get(ATTRIBUTE_DEVICE)
	' Log($"Devices found: ${Statelist.Size}"$)	'Devicesfound: 14

	' Loop over the statelist to get the device as map and add to the devices selection list (delivelist used for the customlistview devices)
	' The devicelist contains an object with an array holding the two attributes device name and device ise_id
	' To get data from the list (l) use a for each loop like:
	'	For Each o() As Object In l
	'		Dim objName As String = o(0)
	'		Dim objIseId As String = o(1)
	'	Next
	For Each device As Map In Statelist
		' Get the attributes for a device as a map using "Attribute"
		' Example: Attributes={name=Bad, ise_id=1786, unreach=False, config_pending=False},
		Dim attributes As Map = device.Get(KEY_ATTRIBUTES)
		DeviceList.Add(Array(attributes.Get(ATTRIBUTE_NAME), attributes.Get(ATTRIBUTE_ISE_ID)))
	Next
End Sub
#End Region

#Region PARSESYSVARLIST
' Parse the sysvarlist XML to a MAP and build the devicelist with 2 items name & ised_id
' Retrieve the HomeMatic CCU sysvarlist via XML-API URL request.
' Example: http://192.168.1.70/addons/xmlapi/sysvarlist.cgi or http://ccu3-webui/addons/xmlapi/sysvarlist.cgi
' The result is an XML tree with all sysvars
' Parsed are the attributes: name, ise_id, value
'<systemVariables>
'	<systemVariable name="${sysVarAlarmZone1}" variable="4" value="" value_list="" ise_id="1236" Min="" Max="" unit="" Type="2" subtype="6" logged="false" visible="true" timestamp="0" value_name_0="${sysVarAlarmZone1NotTriggered}" value_name_1="${sysVarAlarmZone1Triggered}"/>
'	<systemVariable name="${sysVarPresence}" variable="1" value="true" value_list="" ise_id="950" Min="" Max="" unit="" Type="2" subtype="2" logged="false" visible="true" timestamp="1613571859" value_name_0="${sysVarPresenceNotPresent}" value_name_1="${sysVarPresencePresent}"/>
'	<systemVariable name="BatteryCheck" variable="OK 25.02 00:00" value="OK 25.02 00:00" value_list="" ise_id="2799" Min="" Max="" unit="" Type="20" subtype="11" logged="false" visible="true" timestamp="1613636206" value_name_0="" value_name_1=""/>
'	<systemVariable name="DutyCycle" variable="2.000000" value="2.000000" value_list="" ise_id="3032" Min="-1" Max="100" unit="%" Type="4" subtype="0" logged="false" visible="true" timestamp="1614176580" value_name_0="" value_name_1=""/>
'	</systemVariables>
' ...
Public Sub Parse_SysvarList(xmlData As String, xmlSave As Boolean)	'ignore
	Dim xm As Xml2Map			'Lib Xml2Map
	Dim ParsedData As Map
	' Option to save the statelist to a textfile in the dirapp folder
	If xmlSave Then
		File.WriteString(File.DirApp, FILE_SYSVARLIST, xmlData.Replace(">", $">${CRLF}"$))
	End If
	' Clear the lists holding the objects
	SysvarList.Clear
	' Init the object used to parse the xml data (string)
	xm.Initialize
	' Parse the data to a map
	ParsedData = xm.Parse(xmlData)
	' Get the systemVariables attribute (ensure the right case) as a map. It holds a single entry "systemVariables"
	Dim mapSysvarList As Map = ParsedData.Get(ATTRIBUTE_SYSTEMVARIABLES)
	' Check if the map contains data
	If Not(mapSysvarList.IsInitialized) Then
		Return
	End If

	' OPTION
	' Save to file to analyse map structure - only required during development
	File.WriteMap(File.DirApp, "sysvarlist.map", mapSysvarList)

	' Get the systemvariables as list from the map sysvarlist single entry. Each device is a map in the sysvarlist.
	' device= [{Attributes={name=Bad, ise_id=1786, unreach=False, config_pending=False},
	Dim SysvarListTemp As List = mapSysvarList.Get(ATTRIBUTE_SYSTEMVARIABLE)
	' Log($"Sysvars found: ${SysvarList.Size}"$)	'Sysvars found: 4

	' Loop over the sysvarlist to get the device as map and add to the sysvar selection list
	' The sysvarlist contains an object with an array holding the 3 attributes name, ise_id, value
	' To get data from the list (l) use a for each loop like:
	'	For Each o() As Object In l
	'		Dim objName As String = o(0)
	'		Dim objIseId As String = o(1)
	'		Dim objValue As String = o(2)
	'	Next
	For Each sysvar As Map In SysvarListTemp
		' Get the attributes for a device as a map using "Attribute"
		' Example: Attributes={name=Bad, ise_id=1786, unreach=False, config_pending=False},
		Dim attributes As Map = sysvar.Get(KEY_ATTRIBUTES)
		SysvarList.Add(Array(attributes.Get(ATTRIBUTE_NAME), attributes.Get(ATTRIBUTE_VALUE), attributes.Get(ATTRIBUTE_ISE_ID)))
	Next
End Sub
#End Region

